// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.

#ifndef OPENCV_VIDEO_LEGACY_CONSTANTS_H
#define OPENCV_VIDEO_LEGACY_CONSTANTS_H

enum
{
    CV_LKFLOW_PYR_A_READY = 1,
    CV_LKFLOW_PYR_B_READY = 2,
    CV_LKFLOW_INITIAL_GUESSES = 4,
    CV_LKFLOW_GET_MIN_EIGENVALS = 8
};

#endif // OPENCV_VIDEO_LEGACY_CONSTANTS_H
